;-------------------------------------------------------------------------------
; DISPOSITIVO I2C SLAVE PARA SERVIR PORTAS ANALOGICAS                 24/11/2024
;
; STATUS: FUNCIONA
;
; Este projeto fornece, para processadores que nao possuem portas analogicas 
; como o Raspberry Pi, acesso simultaneo as portas analogicas AN0, AN1, AN2 e 
; AN3.
;
; Esta versao nao usa LCD.
;
; O PIC usara o modulo MSSP configurado como um dispositivo I2C slave.
;
; As portas analogicas AN0, AN1, AN2 e AN3 serao lidas e os resultados da
; conversao serao armazenados em grupos de 2 bytes binarios nas localizacoes: 
;    RXBuffer+1(2) --> AN0
;    RXBuffer+3(2) --> AN1
;    RXBuffer+5(2) --> AN2
;    RXBuffer+7(2) --> AN3
;
;-------------------------------------------------------------------------------

; ------------------------------------------------------------------------------
; NOTAS:
;
; 1 - O PIC sera alimentado com 5V e, consequentemente, os pinos SCL e SDA do
;     PIC funcionarao com essa tensao. Apesar do Raspberry Pi operar a 3,3V, 
;     ele consegue lidar com essa tensao em seus pinos SCL e SDA. Esses pinos
;     foram conectados diretamente.
;
; 2 - O buffer do I2C e' limitado a 32 bytes pelas bibliotecas SMBUS do Raspberry.
;
; 3 - O primeiro byte do Buffer e' usado para receber um comando do Master.
;
; 4 - O endereco I2C esta configurado no hardware do PIC nos pinos PORTD<7:6:5>.
;
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
; 2008 Microchip Technology Inc. DS00734B-page 13 - AN734
; APPENDIX B: EXAMPLE SLAVE I2C SOURCE CODE (MODIFIED FOR NEWER PIC18 DEVICES)
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
; Este projeto foi adaptado para o microcontrolador PIC18F4520 e foi retirado do
; documento PDF:
;
; "AN734 - 2008 Microchip Technology Inc. DS00734B-page 13
; APPENDIX B: EXAMPLE SLAVE I2C SOURCE CODE (MODIFIED FOR NEWER PIC18 DEVICES)"
;
; Endereco I2C deste dispositivo:
; ------------------------------
; Os 4 bits de mais alta ordem do endereco I2C deste dispositivo e' dado pelo 
; valor do simbolo 'I2C_MSB_ADDR', enquanto os 3 bits de mais baixa ordem sao
; dados pelos bits PORTD<7,6,5>. Assim, os enderecos poderao variar de 0x40 a
; 0x47.
;
;-------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
; HARDWARE
;
; VDD ............ 5V
; Fosc ........... 4MHZ - Cristal 
; RC3 ............ I2C SCL                      IN/OUT
; RC4 ............ I2C SDA                      IN/OUT
; RD1 ............ LED - Programa rodando       OUT
; RD2 ............ LED - Interrupcao            OUT
; RD5 ............ Endereco I2C - LSB           IN
; RD6 ............ Endereco I2C - LSB           IN
; RD7 ............ Endereco I2C - LSB           IN
;
;-------------------------------------------------------------------------------

; ------------------------------------------------------------------------------
; PORTAS ANALOGICAS
; ------------------------------------------------------------------------------
; PINO  NOME
;  2    RA0/AN0 ........ Porta alnalogica             IN
;  3    RA1/AN1 ........ Porta alnalogica             IN
;  4    RA2/AN2 ........ Porta alnalogica             IN
;  5    RA3/AN3 ........ Porta alnalogica             IN
;  7    RA5/AN4 ........ Porta alnalogica             IN
;  8    RE0/AN5 ........ Porta alnalogica             IN
;  9    RE1/AN6 ........ Porta alnalogica             IN
; 10    RE2/AN7 ........ Porta alnalogica             IN
; 35    RB2/AN8 ........ Porta alnalogica             IN
; 36    RB3/AN9 ........ Porta alnalogica             IN
; 34    RB1/AN10 ....... Porta alnalogica             IN
; 37    RB4/AN11 ....... Porta alnalogica             IN
; 33    RB0/AN12 ....... Porta alnalogica             IN
;
; ------------------------------------------------------------------------------

;-------------------------------------------------------------------------------
;Exemplo de programa Python3 para testar este protocolo no Raspberry Pi 5
;
; Notas:
; 1 - O buffer do I2C e' limitado a 32 bytes pela biblioteca smbus.
;
;-------------------------------------------------------------------------------
;import subprocess
;from time import sleep
;import smbus
;I2C_ADDRESS = 0x40
;bus = smbus.SMBus(1)
;lm35_step = .48876      # Volts/Grau 
;	
;while True:
;   value = bus.read_i2c_block_data(I2C_ADDRESS, 0, 8)
;   t = ( ( 256 * value[5] ) + value[6] ) * lm35_step
;   tf = f"{t:.2f}"
;   print("TEMPERATURA: " + tf, end="\r")
;	sleep(1)	
;
;-------------------------------------------------------------------------------




;-------------------------------------------------------------------------------
; Include Files
;-------------------------------------------------------------------------------
	list      p=18F4520
    #include <p18F4520.inc>
	errorlevel  -302  

; ------------------------------------------------------------------------------
; CONFIGURATION BITS ou FUSES
; ------------------------------------------------------------------------------
    CONFIG WDT    = OFF
	CONFIG MCLRE  = ON
	CONFIG DEBUG  = OFF
	CONFIG LVP    = OFF 
;    CONFIG OSC    = INTIO67      ; oscilador interno (pode ser tambem INTIO7)
;    CONFIG OSC    = HSPLL        ; oscilador com cristal de 10MHz com PLL (4 x Fosc1) fica 40MHz
    CONFIG OSC    = HS            ; oscilador a cristal externo

               #include  "../../Includes/familias.inc"

; ------------------------------------------------------------------------------
; Macros auxiliares
; ------------------------------------------------------------------------------
memset         macro   Buf_addr,Value,Length
               movlw   Length                    ; This macro loads a range of data memory
               movwf   Temp                      ; with a specified value. The starting
               movlw   Buf_addr                  ; address and number of bytes are also
               movwf   FSR0L                     ; specified.
SetNext        movlw   Value
               movwf   INDF0
               incf    FSR0L,F
               decfsz  Temp, F
               goto    SetNext
               endm

load           macro   Address,Offset            ; This macro loads the correct value
               movlw   Address                   ; into the FSR given an initial data
               movwf   FSR0L                     ; memory address and offset value.
               movf    Offset,W
               addwf   FSR0L,F
               endm

; ------------------------------------------------------------------------------
; Simbolos
; ------------------------------------------------------------------------------
I2C_MSB_ADDR   equ     0x40
CLOCK          equ     .4000000
#define        RX_BUF_LEN    .32       ; MAXIMO 32 Bytes

; ------------------------------------------------------------------------------
; RAM
; ------------------------------------------------------------------------------
               UDATA   0x80
RXBuffer       res     RX_BUF_LEN      ; MAXIMO 32 Bytes (manter neste endereco)
                                       ; Limitado a 32 bytes pela biblioteca
                                       ; 'smbus' do Python no Raspberry Pi   

WRK1           res     1
WRK0           res     1

temp_temp      res     1
delay_temp     res     1
addr7bits      res     1

svw            res     1
svstatus       res     1
FSRsave        res     1
PCLATHsave     res     1
Index          res     1
Temp           res     1

            #include  "../../Includes/gpr.inc"

;-------------------------------------------------------------------------------
; Power On/Reset
;-------------------------------------------------------------------------------
PRG            CODE    0x00
               goto    inicio

INT1           CODE    0x04
               retfie

INT2           CODE    0x08
               goto    FLIH

MAIN           CODE

               #include  "../../Services/delay_services.inc"

;-------------------------------------------------------------------------------
; Inicio do programa
;-------------------------------------------------------------------------------
inicio         movlw   0x0F                      ;  
               movwf   ADCON1                    ; todas as portas sao I/O digital

               clrf    Index  
               clrf    Temp 
               clrf    RXBuffer

               movlw   0
               movwf   TRISA
               movwf   TRISB
               movwf   TRISC
               movwf   TRISD
               movwf   TRISE
               movwf   PORTA
               movwf   PORTB
               movwf   PORTC
               movwf   PORTD
               movwf   PORTE

; ------------------------------------------------------------------------------
; Configuracao do dispositivo I2C
; ------------------------------------------------------------------------------
               call    i2cSetup

; ------------------------------------------------------------------------------
; Ronda principal
; ------------------------------------------------------------------------------
loop           btg     LATD, 1

               movlw   .250
               call    delay_ms
               movlw   .250
               call    delay_ms

               goto    loop

; ------------------------------------------------------------------------------ 
; Leitura das portas analogicas
;
; As portas analogicas AN0, AN1, AN2 e AN3 serao lidas e os resultados da
; conversao, apos serem traduzidos para ASCII, serao armazenados em RXBuffer+1,
; RXBuffer+5, RXBuffer+9 e RXBuffer+13 respectivamente.
;
; ------------------------------------------------------------------------------
adcConvert
         clrf     ADCON0

         ; Configura ADCON1 --> Seleciona as portas analogicas AN3, AN2, AN1 e AN0
         movlw    b'00001011'                        ; AN3, AN2, AN1 e AN0
         movwf    ADCON1

         ; Configura portas de I/O
         movlw    b'00001111'                        ; configura como entradas
         movwf    TRISA

          ; Configura ADCON2
          movlw   b'00010100'                        ; bits<5:3> --> 4 Tad     bits<3:0> --> Fosc/4               
          movwf   ADCON2
          bsf     ADCON2,ADFM                        ; alinhar a direita

          ; Converte canal 0 - AN0
          movlw   b'00000000'                        ; numero do canal 0
          movwf   ADCON0                             ; seleciona o canal
          call    adcConv                            ; leitura do canal
          lfsr    FSR1, RXBuffer+.1                  ; aponta local no buffer
          call    toString                           ; deposita o resultado no buffer

          ; Converte canal 1 - AN1
          movlw   b'00000100'                        ; numero do canal 1
          movwf   ADCON0                             ; seleciona o canal
          call    adcConv                            ; leitura do canal
          lfsr    FSR1, RXBuffer+.3                  ; aponta local no buffer
          call    toString                           ; deposita o resultado no buffer

          ; Converte canal 2 - AN2
          movlw   b'00001000'                        ; numero do canal 2
          movwf   ADCON0                             ; seleciona o canal
          call    adcConv                            ; leitura do canal
          lfsr    FSR1, RXBuffer+.5                  ; aponta local no buffer
          call    toString                           ; deposita o resultado no buffer

          ; Converte canal 3 - AN3
          movlw   b'00001100'                        ; numero do canal 3
          movwf   ADCON0                             ; seleciona o canal
          call    adcConv                            ; leitura do canal
          lfsr    FSR1, RXBuffer+.7                  ; aponta local no buffer
          call    toString                           ; deposita o resultado no buffer

          movlw   0                                  ; marcador de fim do buffer
          movwf   POSTINC1                           ; marca o fim do buffer
          movwf   POSTINC1                           ; marca o fim do buffer

          return

adcConv   bsf     ADCON0,ADON                        ; habilita ADC
          bsf     ADCON0,GO_DONE                     ; iniciar conversao
convwait  btfsc   ADCON0,GO_DONE                     ; terminou???
          goto    convwait                           ; NAO --> espera conversao terminar
          movf    ADRESH,w                           ; obtem o MSB da conversao
          andlw   0x03                               ; deixa so bits 1 e 0 do MSB
          movwf   WRK1                               ; armazena valor convertido
          movff   ADRESL,WRK0                        ; armazena LSB do valor convertido
          return

toString  movff    WRK1, POSTINC1                    ; armazena MSB do byte convertido no buffer
          movff    WRK0, POSTINC1                    ; armazena LSB do byte convertido no buffer
          return

;-------------------------------------------------------------------------------
; Configuracoes iniciais do I2C
;-------------------------------------------------------------------------------
i2cSetup       bsf     TRISC, 3                  ; I2C SCL - entrada
               bsf     TRISC, 4                  ; I2C SDA - entrada
               movlw   0x80                      ; UDATA --> 0x80 (endereco MSB do buffer)
               movwf   FSR0H                     ; carrega apontador do buffer
               clrf    FSR0L                     ; LSB = 0

               call    getRealAddress            ; SSPADD --> I2C Address
               movlw   0x36                      ;
               movwf   SSPCON1                   ; SSPCON1 = 0x36
               clrf    SSPSTAT                   ; SSPSTAT = 0x00
               clrf    SSPCON2                   ; SSPCON2 = 0x00
               bsf     SSPCON2,SEN               ; habilita Clock Stretching
               bcf     PIR1, SSPIF               ; limpa flag de interrupcao MSSP
               bsf     PIE1, SSPIE               ; habilita interrupcao MSSP
               bsf     INTCON, PEIE              ; habilita interrupcoes de perifericos
               bsf     INTCON, GIE               ; habilita interrupcoes globais
               return                            ; fim

;-------------------------------------------------------------------------------
; Interrupcao - Primeiro nivel
;-------------------------------------------------------------------------------
FLIH           btfsc   PIR1, SSPIF               ; interrupcao do SSP?
               goto    isrSSP                    ; SIM --> atende 
               retfie                            ; NAO --> ignora

;-------------------------------------------------------------------------------
; ISR - Rotina de interrupcao SSP
;-------------------------------------------------------------------------------
isrSSP         bsf     LATD, 2                   ; acende LED indicador de interrupcao

               movwf   svw                       ; salva W (MSB + LSB)
               swapf   STATUS,w                  ; obtem o STATUS
               movwf   svstatus                  ; salva o STATUS antes da interrupcao

               movf    FSR0L, W                  ;
               movwf   FSRsave                   ; Save FSR

               call    SSP_Handler               ; leitura das portas analogicas

               movf    FSRsave, W                ;
               movwf   FSR0L                     ; Restaura FSR

               swapf   svstatus,w                ; recupera STATUS
               movwf   STATUS                    ; restaura o STATUS antes da interrupcao
               swapf   svw,f                     ; inverte nibles (LSB + MSB)
               swapf   svw,w                     ; restaura W sem afetar o STATUS

               bsf     SSPCON1, CKP              ; libera o clock (clock stretching)
               bcf     LATD, 2                   ; apaga LED indicador de interrupcao
               bcf     PIR1, SSPIF               ; limpa o flag de interrupcao
               retfie  FAST                      ; retorna da interrupcao

; ------------------------------------------------------------------------------
; Obtem o endereco real I2C                                           08/11/2024
;
; Carrega o registrador SSPADD com o endereco de 8 bits, usando o valor
; lido nos pinos RD7:RD6:RD5.
; O Nible de mais alta ordem sera fixo e dado pelo valor da constante
; I2C_MSB_ADDRESS.
; ------------------------------------------------------------------------------
getRealAddress
          movlw b'11100000'        ; pinos RD7:RD6:RD5 --> entrada
          movwf TRISD              ; configura entradas
          swapf PORTD, w           ; le o PORTD invertendo os nibles (RD3:RD2:RD1:RD: RD7:RD6:RD5:RD4)
          movwf R1                 ; salva a leitura (RD3:RD2:RD1:RD: RD7:RD6:RD5:RD4)
          rrcf  R1, f              ; posiciona os 3 bits de endereco (0:3:RD2:RD1:RD: RD7:RD6:RD5)
          movlw 0x07               ; mascara de limpeza
          andwf R1, f              ; deixa somente os bits de mais baixa ordem (0:0:0:0:0:RD7:RD6:RD5)
          movlw I2C_MSB_ADDR       ; parte fixa do endereco
          iorwf R1, w              ; forma o endereco real de 7 bits
          movwf addr7bits          ; salva endereco de 7 bits
          rlcf  addr7bits, w       ; forma o endereco de 8 bits
          movwf SSPADD             ; carrega o registrador de endereco SSP/I2C
          return

;-------------------------------------------------------------------------------
; ROTINA DE SERVICO DE INTERRUPCAO
;
; SSPSTAT: SMP  CKE   D/A   P   S   R/W   UA   BF
;           7    6     5    4   3    2     1    0
;
; A cada entrada na interrupcao, o registrador SSPSTAT representara um dos
; estados abaixo que sera usado para atender a requisicao do Master.
;
; State 1: Requisicao de gravacao - O ultimo byte era um endereco
; SSPSTAT bits: S = 1, D_A = 0, R_W = 0, BF = 1
;                SMP  CKE   D/A   P   S   R/W   UA   BF
;                 x    x     0    x   1    0     x    1
;
; State 2: Requisicao de gravacao - O ultimo byte era um dado
; SSPSTAT bits: S = 1, D_A = 1, R_W = 0, BF = 1
;                SMP  CKE   D/A   P   S   R/W   UA   BF
;                 x    x     1    x   1    0     x    1
;
; State 3: Requisicao de leitura - O ultimo byte era um endereco (Le o primeiro dado)
; SSPSTAT bits: S = 1, D_A = 0, R_W = 1 (see Appendix C for more information)
;                SMP  CKE   D/A   P   S   R/W   UA   BF
;                 x    x     0    x   1    1     x    x
;
; State 4: Requisicao de leitura - O ultimo byte era um dado (Le o proximo dado)
; SSPSTAT bits: S = 1, D_A = 1, R_W = 1, BF = 0
;                SMP  CKE   D/A   P   S   R/W   UA   BF
;                 x    x     1    x   1    1     x    0
;
; State 5: Recebeu um NACK do Master
; SSPSTAT bits: S = 1, D_A = 1, BF = 0, SSPCON<CKP> = 1 (ver Appendix C)
;                SMP  CKE   D/A   P   S   R/W   UA   BF
;                 x    x     1    x   1    x     x    0
;
;-------------------------------------------------------------------------------
SSP_Handler    movf    SSPSTAT, W                ; pega o SSPSTAT
               andlw   b'00101101'               ; deixa so o que interessa em SSPSTAT (D/A, S, R/W e BF)
               movwf   Temp                      ; salva

State1:        ; Requisicao de gravacao: O ultimo byte era um endereco
               ; SSPSTAT: | SMP | CKE | D/A | P |  S | R/W | UA | BF |
               ; Quando este estado for detectada:
               ; 1 - Aponta o indice para o primeiro byte do buffer RXBuffer
               ; 2 - Le e despreza o endereco em SSPBUF
               movlw   b'00001001'               ; address, buffer is full.  D/A=0  S=1 BF=1
               xorwf   Temp,W                    ;
               btfss   STATUS,Z                  ; tem endereco no SSPBUF???
               goto    State2                    ; NAO --> verifica se estamos em outro estado
               clrf    Index                     ; SIM --> zera o indice
               movf    SSPBUF,W                  ; despreza endereco em SSPBUF.
               return                            ; fim

State2:        ; Requisicao de gravacao: O ultimo byte era um dado
               ; SSPSTAT: | SMP | CKE | D/A | P |  S | R/W | UA | BF |
               ; Quando este estado for detectado:
               ; 1 - Aponta FSR0H:FSR0L para o deslocamento do indice
               ; 2 - Le o valor presente em SSPBUF
               ; 3 - Armazena o valor em RXBuffer na posicao corrente do indice
               ; 4 - Avanca o indice de 1 byte
               ; 5 - Se index==RX_BUFF_LEN --> index=0 
               movlw   b'00101001'               ; D/A=1  S=1  BF=1
               xorwf   Temp,W
               btfss   STATUS,Z                  ; estamos no estado esperado?
               goto    State3                    ; NAO --> ver proximo estado
               load    RXBuffer,Index            ; aponta para o buffer de leitura
               movf    SSPBUF,W                  ; le o byte recebido
               movwf   INDF0                     ; coloca o byte no buffer de leitura
               incf    Index,F                   ; proximo byte no buffer de leitura
               movf    Index,W                   ; pega o indice corrente
               sublw   RX_BUF_LEN                ; verifica
               btfsc   STATUS,Z                  ; estouro do buffer de leitura?
               clrf    Index                     ; SIM --> reposiciona no inicio
               call    adcConvert                ; NAO --> realiza conversao A/D
               return                            ; fim

State3:        ; Requisicao de leitura: O ultimo byte era um endereco (Le o primeiro dado)
               ; SSPSTAT: | SMP | CKE | D/A | P |  S | R/W | UA | BF |
               ; Quando este estado for detectado:
               ; 1 - Le e despreza o dado (address) em SSPBUF
               ; 2 - A ponta o indice para o primeiro byte do buffer RXBuffer
               ; 3 - Faz FSR0H:FSR0L apontar o inicio do buffer RXBuffer
               ; 4 - Envia ao Master o primeiro byte do buffer RXBuffer
               ; 5 - Avanca o indice para o segundo byte do RXBuffer
               movf    Temp,W                    ;
               andlw   b'00101100'               ; deixa so os bits que interessam
               xorlw   b'00001100'               ; if( D/A=0 && S=1 && R/W=1) --> requer leitura???
               btfss   STATUS, Z                 ; SIM --> vai ler
               goto    State4                    ; NAO --> verifica proximo estado
               movf    SSPBUF, W                 ; le e despreza o dado
               clrf    Index                     ; indice aponta o inicio do buffer
               load    RXBuffer, Index           ; FSR0H:FSR0L -> aponta inicio do buffer
               movf    INDF0, W                  ; pega o byte do inicio do buffer.
               call    WriteI2C                  ; envia para o Master
               incf    Index, F                  ; avanca o indice
               return                            ; fim

State4         ; Requisicao de gravacao: O ultimo byte era um dado (Le o proximo byte)
               ; SSPSTAT: | SMP | CKE | D/A | P |  S | R/W | UA | BF |
               ; Quando este estado for detectado:
               ; 1 - Verifica se esta em "clock Stretch"
               ; 2 - Verifica se SSPSTAT<BF>=0 (buffer vazio)
               ; 3 - Zera o indice se estourou (index==RX_BUF_LEN)
               ; 4 - FSR0H:FSR0L aponta para o byte corrente do indice
               ; 5 - Envia ao Master o byte corrente no RXBuffer
               ; 6 - Avanca o indice 1 byte
               btfsc   SSPCON1, CKP              ; esta em "clock stretch"???
               goto    State5                    ; NAO --> proximo estado
               movlw   b'00101100'               ; SIM --> 
               xorwf   Temp, W                   ; somente limpa o flag BF
               btfss   STATUS, Z                 ; estamos no estado esperado?
               goto    State5                    ; NAO --> verifica proximo estado
               movf    Index, W                  ; SIM --> o indice
               sublw   RX_BUF_LEN                ; verifica estouro do buffer
               btfsc   STATUS, Z                 ; passou dos limites do buffer?
               clrf    Index                     ; SIM --> reposiciona no inicio do buffer
               load    RXBuffer, Index           ; NAO --> posiciona no byte corrente
               movf    INDF0, W                  ; pega o byte corrente no buffer
               call    WriteI2C                  ; envia ao Master
               incf    Index, F                  ; posiciona o proximo byte no buffer
               return                            ; fim

State5         ; Recebeu um NACK do Master - Simplesmente retorna
               ; SSPSTAT: | SMP | CKE | D/A | P |  S | R/W | UA | BF
               nop                               ; 
               return                            ; fim

;-------------------------------------------------------------------------------
; WriteI2C
; W contem o dado a enviar
;-------------------------------------------------------------------------------
WriteI2C       btfsc   SSPSTAT ,BF               ; SSPBUF esta cheio???
               goto    WriteI2C                  ; SIM --> espera esvaziar
DoI2CWrite     bcf     SSPCON1, WCOL             ; NAO --> limpa flag de colisao
               movwf   SSPBUF                    ; envia o dado para o Master
               btfsc   SSPCON1, WCOL             ; ocorreu uma colisao???
               goto    DoI2CWrite                ; SIM --> tenta novamente
               return                            ; NAO --> OK

 
               END


